#include <itkBilateralImageFilter.h>
#include <vtkITKBilateralImageFilter.h>

#include <functional>

#include <ciso646>  // detect std::lib

#ifdef _WIN32
 #include <type_traits>
#else
 #ifdef _LIBCPP_VERSION
 // using libc++
 #include <type_traits>
 #else
 // using libstdc++
 #include <tr1/type_traits>
#endif

#endif

#include <vtkDataArray.h>
#include <vtkObjectFactory.h>
#include <vtkImageData.h>
#include <vtkInformationVector.h>
#include <vtkInformation.h>
#include <vtkPointData.h>
#include <vtkStreamingDemandDrivenPipeline.h>

#include <itkImageToVTKImageFilter.h>
#include <itkVTKImageToImageFilter.h>

typedef unsigned char uint8_t;
typedef unsigned int  uint32_t;

#define ITK_UCHAR_TYPE
#define ITK_CHAR_TYPE
#define ITK_USHORT_TYPE
#define ITK_SHORT_TYPE
#define ITK_UINT_TYPE
#define ITK_INT_TYPE
#define ITK_ULONG_TYPE
#define ITK_LONG_TYPE
#define ITK_FLOAT_TYPE
#define ITK_DOUBLE_TYPE

vtkStandardNewMacro(vtkITKBilateralImageFilter);

vtkITKBilateralImageFilter::vtkITKBilateralImageFilter()
{
  this->SetNumberOfInputPorts(1);
  this->SetNumberOfOutputPorts(1);

  this->DomainSigma = 4.0;
  this->RangeSigma = 50.0;
  this->NumberOfRangeGaussianSamples = 100u;
  this->SetInputArrayToProcess(0,0,0,vtkDataObject::FIELD_ASSOCIATION_POINTS,
    vtkDataSetAttributes::SCALARS);
}

vtkITKBilateralImageFilter::~vtkITKBilateralImageFilter()
{
}

int vtkITKBilateralImageFilter::RequestUpdateExtent(
  vtkInformation * vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector * vtkNotUsed(outputVector))
{
  int inWExt[6];
  vtkInformation *inInfo0 = inputVector[0]->GetInformationObject(0);
  inInfo0->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), inWExt);
  inInfo0->Set(vtkStreamingDemandDrivenPipeline::UPDATE_EXTENT(), inWExt, 6);

  return 1;
}

int vtkITKBilateralImageFilter::FillInputPortInformation(int port, vtkInformation * info)
{
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkImageData");
  info->Set(vtkAlgorithm::INPUT_IS_REPEATABLE(), 0);
  info->Set(vtkAlgorithm::INPUT_IS_OPTIONAL(), 0);

  if (port == 0)
    {
    return 1;
    }

  return 0;
}

int vtkITKBilateralImageFilter::RequestInformation(vtkInformation *vtkNotUsed(request), vtkInformationVector **inputVector, vtkInformationVector *outputVector)
{
  double spacing[3];
  double origin[3];
  int wholeExt[6];

  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  inInfo->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), wholeExt);
  inInfo->Get(vtkDataObject::SPACING(), spacing);
  inInfo->Get(vtkDataObject::ORIGIN(), origin);

  outInfo->Set(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), wholeExt, 6);
  outInfo->Set(vtkDataObject::SPACING(), spacing, 3);
  outInfo->Set(vtkDataObject::ORIGIN(), origin, 3);

  return 1;
}

int vtkITKBilateralImageFilter::RequestData(vtkInformation * vtkNotUsed(info), vtkInformationVector **inputVector, vtkInformationVector *outputVector)
{
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkImageData *originalInput = vtkImageData::SafeDownCast
    (inInfo->Get(vtkDataObject::DATA_OBJECT()));
  if (!originalInput)
    {
    vtkErrorMacro("Input is not of type vtkImageData");
    return 0;
    }
  
  vtkImageData *input = originalInput;
  
  vtkInformation *outInfo = outputVector->GetInformationObject(0);
  vtkImageData *output = vtkImageData::SafeDownCast
    (outInfo->Get(vtkDataObject::DATA_OBJECT()));
  if (!output)
    {
    vtkErrorMacro("Output is not of type vtkImageData");
    return 0;
    }
  
  // Set the active scalar array in the inputs. Normally, this would be
  // done in the filter that imports the VTK image into ITK, but that
  // filter doesn't expose a way to set the active scalar in the interface.
  char * activeArrayName0 = this->GetInputArrayToProcess( 0, input )->GetName();
  input->GetPointData()->SetActiveScalars( activeArrayName0 );
  
  bool success = true;
  switch( input->GetPointData()->GetScalars()->GetDataType() )
    {
  #ifdef ITK_UCHAR_TYPE
    case VTK_UNSIGNED_CHAR:
      success = Run( static_cast< unsigned char * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_CHAR_TYPE
    case VTK_CHAR:
    case VTK_SIGNED_CHAR:
      success = Run( static_cast< char * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_USHORT_TYPE
    case VTK_UNSIGNED_SHORT:
      success = Run( static_cast< unsigned short * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_SHORT_TYPE
    case VTK_SHORT:
      success = Run( static_cast< short * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_UINT_TYPE
    case VTK_UNSIGNED_INT:
      success = Run( static_cast< unsigned int * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_INT_TYPE
    case VTK_INT:
      success = Run( static_cast< int * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_ULONG_TYPE
    case VTK_UNSIGNED_LONG:
      success = Run( static_cast< unsigned long * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_LONG_TYPE
    case VTK_LONG:
      success = Run( static_cast< long * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_FLOAT_TYPE
    case VTK_FLOAT:
      success = Run( static_cast< float * >( 0 ), input, output );
      break;
  #endif
  #ifdef ITK_DOUBLE_TYPE
     case VTK_DOUBLE:
      success = Run( static_cast< double * >( 0 ), input, output );
      break;
  #endif
  
    default:
      vtkErrorMacro( "Unknown pixel type encountered" );
      success = false;
      break;
    }
  return (success ? 1 : 0);
}

template< class TInputPixel >
bool vtkITKBilateralImageFilter::Run(TInputPixel *, vtkImageData * vtkInput0, vtkImageData * output)
{
  typedef itk::Image< TInputPixel, 3 > InputImageType;
  typedef InputImageType OutputImageType;

  typedef itk::BilateralImageFilter< InputImageType,  OutputImageType > FilterType;
  typename FilterType::Pointer filter = FilterType::New();

  typedef itk::VTKImageToImageFilter< InputImageType > VTKToITKConnectorType0;
  typename VTKToITKConnectorType0::Pointer vtkToITKConnector0 = VTKToITKConnectorType0::New();
  vtkToITKConnector0->SetInput( vtkInput0 );
  try
    {
    vtkToITKConnector0->Update();
    }
  catch ( itk::ExceptionObject & except )
    {
    vtkErrorMacro( << except );
    return false;
    }

  InputImageType * image1 = vtkToITKConnector0->GetOutput();
  filter->SetInput( image1 );

  filter->SetDomainSigma(this->GetDomainSigma());
  filter->SetRangeSigma(this->GetRangeSigma());
  filter->SetNumberOfRangeGaussianSamples(this->GetNumberOfRangeGaussianSamples());
  try
    {
    filter->Update();
    }
  catch ( itk::ExceptionObject & except )
    {
    vtkErrorMacro( << except );
    return false;
    }

  typedef itk::ImageToVTKImageFilter< OutputImageType > ITKToVTKConnectorType;
  typename ITKToVTKConnectorType::Pointer itkToVTKConnector = ITKToVTKConnectorType::New();
  itkToVTKConnector->SetInput( filter->GetOutput() );
  itkToVTKConnector->Update();

  output->DeepCopy( itkToVTKConnector->GetOutput() );

  return true;
}

void vtkITKBilateralImageFilter::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);

  os << indent << "DomainSigma:" << this->DomainSigma << "\n";
  os << indent << "RangeSigma:" << this->RangeSigma << "\n";
  os << indent << "NumberOfRangeGaussianSamples:" << this->NumberOfRangeGaussianSamples << "\n";
}
