#ifndef __itkBinaryBranchCountingImageFilter_h
#define __itkBinaryBranchCountingImageFilter_h

#include "itkImageToImageFilter.h"
#include "itkImageRegionIteratorWithIndex.h"
#include "itkNeighborhoodIterator.h"
#include <fstream>  
#include <stdio.h>  
#include <iostream>  
#include <string.h>  
#include <string>  
#include <vector>  
#include <stack>  
#include <deque>  
#include <queue>  
#include <map>  
#include <math.h>  
#include <algorithm>  
#include <numeric>  
#include <functional>  
#include <memory.h>  

using namespace std;
namespace itk
{
	/** \class BinaryBranchCountingImageFilter
	* Author: Xiaojie Zhao
	*
	* \brief This filter count branch numbers for the thinning
	* result (using BinaryThinningImageFilter3D method in image surfer) 
	* of the microscopy
	* fiber image.
	* 
	* This is a 4-step method. 
	* 1) Count the 26-neighbors for each point. All the points that have more than 2 neighbors are called fake branch points. 
	* 2) For every fake branch point, count how many 2-branch (regular) neighbors it has. (See the red numbers in the following figure.
	* 3) Group those fake branch points which are in a small neighborhood of each other. Tag the sum of their 
	branch numbers to one of these points. This point is then considered the real branch point. All others are labeled as regular points.
	* 4) Combine 3-branch points that are close to each other to a 4 branch points. (This is a special processing for the microscopy fibrin image.
	*/

	template< class TInputImage, class TOutputImage >
	class ITK_EXPORT BinaryBranchCountingImageFilter:
		public ImageToImageFilter< TInputImage, TOutputImage >
	{
	public: 

		/** Standard class typedefs. */
		typedef BinaryBranchCountingImageFilter                        Self;
		typedef ImageToImageFilter< TInputImage, TOutputImage > Superclass;
		typedef SmartPointer< Self >                            Pointer;
		typedef SmartPointer< const Self >                      ConstPointer;

		/** Method for creation through the object factory */
		itkNewMacro(Self);

		/** Run-time type information (and related methods). */
		itkTypeMacro(BinaryBranchCountingImageFilter, ImageToImageFilter);

		/** Type for input image. */
		typedef   TInputImage InputImageType;

		/** Type for output image: Skelenton of the object.  */
		typedef   TOutputImage OutputImageType;

		/** Type for the region of the input image. */
		typedef typename InputImageType::RegionType RegionType;

		/** Type for the index of the input image. */
		typedef typename RegionType::IndexType IndexType;

		/** Type for the index of the input image. */
		typedef typename InputImageType::PixelType PixelType;

		/** Type for the size of the input image. */
		typedef typename RegionType::SizeType SizeType;

		/** Pointer Type for input image. */
		typedef typename InputImageType::ConstPointer InputImagePointer;

		/** Pointer Type for the output image. */
		typedef typename OutputImageType::Pointer OutputImagePointer;

		/** Neighborhood iterator type */
		typedef NeighborhoodIterator< TInputImage > NeighborhoodIteratorType;

		/** Get Skelenton by thinning image. */
		OutputImageType * GetPruning(void);

		/** ImageDimension enumeration   */
		itkStaticConstMacro(InputImageDimension, unsigned int,
			TInputImage::ImageDimension);
		itkStaticConstMacro(OutputImageDimension, unsigned int,
			TOutputImage::ImageDimension);


		//xiaojie
		itkSetMacro(HandleBoundaryRadius, int);
		itkGetConstMacro(HandleBoundaryRadius, int);
		itkSetMacro(BranchCombineRadius, int);
		itkGetConstMacro(BranchCombineRadius, int);


#ifdef ITK_USE_CONCEPT_CHECKING
		/** Start concept checking */
		itkConceptMacro( SameDimensionCheck,
			( Concept::SameDimension< InputImageDimension, OutputImageDimension > ) );
		itkConceptMacro( SameTypeCheck,
			( Concept::SameType< PixelType, typename TOutputImage::PixelType > ) );
		itkConceptMacro( AdditiveOperatorsCheck,
			( Concept::AdditiveOperators< PixelType > ) );
		itkConceptMacro( IntConvertibleToPixelTypeCheck,
			( Concept::Convertible< int, PixelType > ) );
		itkConceptMacro( PixelLessThanIntCheck,
			( Concept::LessThanComparable< PixelType, int > ) );
		/** End concept checking */
#endif
	protected:
		BinaryBranchCountingImageFilter();
		virtual ~BinaryBranchCountingImageFilter() {}
		void PrintSelf(std::ostream & os, Indent indent) const;

		/** Compute thinning Image. */
		void GenerateData();

		/** Compute thinning Image. */
		void CountBranches();

	private:
		BinaryBranchCountingImageFilter(const Self &); //purposely not implemented
		void operator=(const Self &);           //purposely not implemented

		bool path() const;
		int m_HandleBoundaryRadius;
		int m_BranchCombineRadius;
	}; // end of BinaryThinningImageFilter class
} //end namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkBinaryBranchCountingImageFilter.hxx"
#endif

#endif
