#ifndef __itkBinaryBranchCountingImageFilter_hxx
#define __itkBinaryBranchCountingImageFilter_hxx

#include <iostream>

#include "itkBinaryBranchCountingImageFilter.h"
#include "itkImageRegionIterator.h"
#include "itkImageDuplicator.h"
#include "itkConstantBoundaryCondition.h"

namespace itk
{
	/**
	*    Constructor
	*/
	template< class TInputImage, class TOutputImage >
	BinaryBranchCountingImageFilter< TInputImage, TOutputImage >
		::BinaryBranchCountingImageFilter()
	{
		this->SetNumberOfRequiredOutputs(1);

		OutputImagePointer branchScalarImage = OutputImageType::New();
		this->SetNthOutput( 0, branchScalarImage.GetPointer() );

	}

	/**
	*  Return the Image pointer
	*/
	template< class TInputImage, class TOutputImage >
	typename BinaryBranchCountingImageFilter<
		TInputImage, TOutputImage >::OutputImageType *
		BinaryBranchCountingImageFilter< TInputImage, TOutputImage >
		::GetPruning(void)
	{
		return dynamic_cast< OutputImageType * >(
			this->ProcessObject::GetOutput(0) );
	}

	/**
	*  Counting branches.
	*/
	template< class TInputImage, class TOutputImage >
	void BinaryBranchCountingImageFilter< TInputImage, TOutputImage >
		::CountBranches(void)
	{
		itkDebugMacro(<< "BranchCounting Start");
		OutputImagePointer branchScalarImage = GetPruning();

		InputImagePointer inputImage  =
			dynamic_cast< const TInputImage  * >( ProcessObject::GetInput(0) );


		branchScalarImage->SetBufferedRegion( branchScalarImage->GetRequestedRegion() );
		branchScalarImage->Allocate();

		typename OutputImageType::RegionType region  = branchScalarImage->GetRequestedRegion();

		ImageRegionConstIterator< TInputImage > it(inputImage,  region);
		ImageRegionIterator< TOutputImage >     ot(branchScalarImage,  region);


		typename NeighborhoodIteratorType::RadiusType radius;
		radius.Fill(1);
		ConstNeighborhoodIterator< InputImageType > itNeighbor(radius, inputImage, region);
		ConstantBoundaryCondition< InputImageType > nbc;
		itNeighbor.OverrideBoundaryCondition(&nbc);

		itNeighbor.GoToBegin();
		it.GoToBegin();
		ot.GoToBegin();

		const unsigned int neighborhoodSize = itNeighbor.Size();
		const unsigned int medianPosition = neighborhoodSize / 2;

		//First round: raw branch counting by searching in 26 neighbors
		while ( !it.IsAtEnd() )
		{
			//// collect all the pixels in the neighborhood, note that we use
			//// GetPixel on the NeighborhoodIterator to honor the boundary conditions
			int numBranch = 0;
			for ( unsigned int i = 0; i < neighborhoodSize; ++i )
			{
				if(itNeighbor.GetPixel(i) > 0.5)
					numBranch ++;
			}

			if(it.Get() < 0.5)
				ot.Set(0);
			else
			{
				ot.Set(numBranch-1);
			}
			++itNeighbor;
			++it;
			++ot;
		}

		//Second round: branch neighbor counting on branch point
		//Create another image to represent whether a voxel is the first regular point near branch

                typedef ImageDuplicator< TOutputImage > DuplicatorType;
		
                typename DuplicatorType::Pointer duplicator1 = DuplicatorType::New();
		duplicator1->SetInputImage(branchScalarImage);
		duplicator1->Update();
		typename TOutputImage::Pointer fiberOnBranch = duplicator1->GetOutput();
		ImageRegionIterator< TOutputImage >   itFiberOnBranch(fiberOnBranch,  region);

		itFiberOnBranch.GoToBegin();
		while(!itFiberOnBranch.IsAtEnd()){
			itFiberOnBranch.Set(0);
			++itFiberOnBranch;
		}

		NeighborhoodIterator< OutputImageType > itOutNeighbor(radius, branchScalarImage, region);
		ConstantBoundaryCondition< OutputImageType > nbcOut;
		itOutNeighbor.OverrideBoundaryCondition(&nbcOut);
		itOutNeighbor.GoToBegin();
		ot.GoToBegin();
		while ( !ot.IsAtEnd() )
		{

			if(abs(2-ot.Get())<1e-6)
			{
				unsigned int i = 0; 
				while ( i < neighborhoodSize )
				{
					if(itOutNeighbor.GetPixel(i)> 2.5)
					{
						itFiberOnBranch.SetIndex(itOutNeighbor.GetIndex(i));
						itFiberOnBranch.Set(itFiberOnBranch.Get()+1);
					}
					i++;
				}
			}

			++ot;
			++itOutNeighbor;
		}


		//Third round: Real Branch Counting
		//Create another image to represent whether a voxel is fiber end on branch

		NeighborhoodIterator< InputImageType > itFiberOnBranchNeighbor(radius, fiberOnBranch, region);
		itFiberOnBranchNeighbor.OverrideBoundaryCondition(&nbcOut);
		ot.GoToBegin();
		itFiberOnBranch.GoToBegin();
		itFiberOnBranchNeighbor.GoToBegin();
		while ( !itFiberOnBranch.IsAtEnd() )
		{
			if(itFiberOnBranch.Get()>0.5)
			{

				unsigned int i = 0;
				while ( i < neighborhoodSize )
				{

					if(itFiberOnBranchNeighbor.GetPixel(i)> 0.5 && i!=medianPosition)
					{
						itFiberOnBranchNeighbor.SetCenterPixel(itFiberOnBranch.Get() 
							+ itFiberOnBranchNeighbor.GetPixel(i));
						itFiberOnBranchNeighbor.SetPixel(i, 0);
						ot.SetIndex(itFiberOnBranchNeighbor.GetIndex(i));
						ot.Set(2);
					}
					i++;
				}

			}

			++it;
			++itFiberOnBranch;
			++itFiberOnBranchNeighbor;
		}
		itFiberOnBranch.GoToBegin();
		ot.GoToBegin();
		while(!ot.IsAtEnd()){
			if(itFiberOnBranch.Get()>0.5)
				ot.Set(itFiberOnBranch.Get());

			else
				itFiberOnBranch.Set(ot.Get());

			++ot;
			++itFiberOnBranch;

		}

		//Fourth round: combine close 3-branch points to 4-branch points

		radius.Fill(m_BranchCombineRadius);
		NeighborhoodIterator< InputImageType > itFiberOnBranchNeighborCombine(radius, fiberOnBranch, region);
		itFiberOnBranchNeighborCombine.OverrideBoundaryCondition(&nbcOut);

		const unsigned int neighborhoodSizeCombine = itFiberOnBranchNeighborCombine.Size();
		const unsigned int medianPositionCombine = neighborhoodSizeCombine / 2;

		ot.GoToBegin();
		itFiberOnBranch.GoToBegin();
		itFiberOnBranchNeighborCombine.GoToBegin();

		while ( !itFiberOnBranch.IsAtEnd() )
		{
			if(abs(itFiberOnBranch.Get()-3)<1e-6)
			{
				int combine = 3;
				unsigned int i = 0;
				while ( i < neighborhoodSizeCombine )
				{

					if(abs(itFiberOnBranchNeighborCombine.GetPixel(i)-3)<1e-6 
						&& i!=medianPositionCombine)
					{
						combine ++;
						itFiberOnBranchNeighborCombine.SetCenterPixel(combine);
						itFiberOnBranchNeighborCombine.SetPixel(i, 2);
					}
					i++;
				}

			}

			++itFiberOnBranch;
			++itFiberOnBranchNeighborCombine;
		}

		itFiberOnBranch.GoToBegin();
		ot.GoToBegin();
		while(!ot.IsAtEnd()){
			ot.Set(itFiberOnBranch.Get());
			++ot;
			++itFiberOnBranch;
		}

		// Handle boundary
		if(m_HandleBoundaryRadius > 1e-6)
		{
			itk::Size<3> imageSize = inputImage->GetLargestPossibleRegion().GetSize();
			itk::Index<3> otIndex; 
			ot.GoToBegin();
			while(!ot.IsAtEnd()){
				otIndex = ot.GetIndex();
				// if it is on boundary
				bool onBoundary = abs((double)otIndex.GetElement(0) - (double)0) < m_HandleBoundaryRadius  
					|| abs((double)otIndex.GetElement(0) - (double)imageSize.GetElement(0) + 1) < m_HandleBoundaryRadius
					|| abs((double)otIndex.GetElement(1) - (double)0) < m_HandleBoundaryRadius  
					|| abs((double)otIndex.GetElement(1) - (double)imageSize.GetElement(1) + 1) < m_HandleBoundaryRadius
					|| abs((double)otIndex.GetElement(2) - (double)0) < m_HandleBoundaryRadius  
					|| abs((double)otIndex.GetElement(2) - (double)imageSize.GetElement(2) + 1) < m_HandleBoundaryRadius;
				// if on boundary and is 1-branch point, label it as regular point
				if(ot.Get()>0.5 && ot.Get()<1.5 && onBoundary)
					ot.Set(2);
				++ot;
			}
		}

		itkDebugMacro(<< "Branch Counting End");
	}

	/**
	*  Generate BranchCounting ScalarField Image
	*/
	template< class TInputImage, class TOutputImage >
	void
		BinaryBranchCountingImageFilter< TInputImage, TOutputImage >
		::GenerateData()
	{
		itkDebugMacro(<< "GenerateData: Counting Branches.");
		this->CountBranches();

	} // end GenerateData()

	/**
	*  Print Self
	*/
	template< class TInputImage, class TOutputImage >
	void
		BinaryBranchCountingImageFilter< TInputImage, TOutputImage >
		::PrintSelf(std::ostream & os, Indent indent) const
	{
		Superclass::PrintSelf(os, indent);

		os << indent << "Counting Branches: " << std::endl;
		os << indent << "BranchCombineRadius: " << m_BranchCombineRadius << std::endl;
		os << indent << "HandleBoundaryRadius: " << m_HandleBoundaryRadius << std::endl;
	}
} // end namespace itk

#endif
