#ifndef __itkBinaryPruningImageFilter3D_h
#define __itkBinaryPruningImageFilter3D_h

#include "itkImageToImageFilter.h"
#include "itkImageRegionIteratorWithIndex.h"
#include "itkNeighborhoodIterator.h"

namespace itk
{
	/** \class BinaryPruningImageFilter3D
	*  Author: Xiaojie Zhao
	*
	* \brief This filter removes "spurs" of less than a certain
	* length in the input image.
	*  
	* This is a modification of BinaryPruningImageFilter in itk.
	* The problem of the itk method is that it removes certain length 
	* of the fiber form all the fiber ends. So not only the spur of less than
	* a certain length is removed, all other fibers are cut by that length. 
	* The resultant skeleton will lead to a bad localization of the real fiber end points then. 
	* The method is modified to eliminate this side effect. Only "spurs" of less than a certain
	* length are removed in the input image.
	*
	* This class is parametrized over the type of the input image
	* and the type of the output image.
	*
	* The input is assumed to be a binary image.
	*
	* (itk method)This filter is a sequential pruning algorithm and known to be computational time
	* dependable of the image size.  The original algorithm is the N-dimensional version
	* of that given for two dimensions in:
	*
	* Rafael C. Gonzales and Richard E. Woods.
	* Digital Image Processing.
	* Addison Wesley, 491-494, (1993).
	*
	* \sa MorphologyImageFilter
	* \sa BinaryErodeImageFilter
	* \sa BinaryDilateImageFilter
	* \sa BinaryThinningImageFilter
	* \ingroup ImageEnhancement MathematicalMorphologyImageFilters
	* \ingroup ITKBinaryMathematicalMorphology
	*/

	template< class TInputImage, class TOutputImage >
	class ITK_EXPORT BinaryPruningImageFilter3D:
		public ImageToImageFilter< TInputImage, TOutputImage >
	{
	public:
		/** Standard class typedefs. */
		typedef BinaryPruningImageFilter3D                        Self;
		typedef ImageToImageFilter< TInputImage, TOutputImage > Superclass;
		typedef SmartPointer< Self >                            Pointer;
		typedef SmartPointer< const Self >                      ConstPointer;

		/** Method for creation through the object factory */
		itkNewMacro(Self);

		/** Run-time type information (and related methods). */
		itkTypeMacro(BinaryPruningImageFilter3D, ImageToImageFilter);

		/** Type for input image. */
		typedef   TInputImage InputImageType;

		/** Type for output image: Skelenton of the object.  */
		typedef   TOutputImage OutputImageType;

		/** Type for the region of the input image. */
		typedef typename InputImageType::RegionType RegionType;

		/** Type for the index of the input image. */
		typedef typename RegionType::IndexType IndexType;

		/** Type for the index of the input image. */
		typedef typename InputImageType::PixelType PixelType;

		/** Type for the size of the input image. */
		typedef typename RegionType::SizeType SizeType;

		/** Pointer Type for input image. */
		typedef typename InputImageType::ConstPointer InputImagePointer;

		/** Pointer Type for the output image. */
		typedef typename OutputImageType::Pointer OutputImagePointer;

		/** Neighborhood iterator type */
		typedef NeighborhoodIterator< TInputImage > NeighborhoodIteratorType;

		/** Get Skelenton by thinning image. */
		OutputImageType * GetPruning(void);

		/** Set/Get the iteration value */
		itkSetMacro(Iteration, unsigned int);
		itkGetConstMacro(Iteration, unsigned int);

		/** ImageDimension enumeration   */
		itkStaticConstMacro(InputImageDimension, unsigned int,
			TInputImage::ImageDimension);
		itkStaticConstMacro(OutputImageDimension, unsigned int,
			TOutputImage::ImageDimension);

#ifdef ITK_USE_CONCEPT_CHECKING
		/** Start concept checking */
		itkConceptMacro( SameDimensionCheck,
			( Concept::SameDimension< InputImageDimension, OutputImageDimension > ) );
		itkConceptMacro( SameTypeCheck,
			( Concept::SameType< PixelType, typename TOutputImage::PixelType > ) );
		itkConceptMacro( AdditiveOperatorsCheck,
			( Concept::AdditiveOperators< PixelType > ) );
		itkConceptMacro( IntConvertibleToPixelTypeCheck,
			( Concept::Convertible< int, PixelType > ) );
		itkConceptMacro( PixelLessThanIntCheck,
			( Concept::LessThanComparable< PixelType, int > ) );
		/** End concept checking */
#endif
	protected:
		BinaryPruningImageFilter3D();
		virtual ~BinaryPruningImageFilter3D() {}
		void PrintSelf(std::ostream & os, Indent indent) const;

		/** Compute thinning Image. */
		void GenerateData();

		/**  Compute thinning Image. */
		void ComputePruneImage();

	private:
		BinaryPruningImageFilter3D(const Self &); //purposely not implemented
		void operator=(const Self &);           //purposely not implemented

		unsigned int m_Iteration;
	}; // end of BinaryThinningImageFilter class
} //end namespace itk

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkBinaryPruningImageFilter3D.hxx"
#endif

#endif
