#ifndef __itkBinaryPruningImageFilter3D_hxx
#define __itkBinaryPruningImageFilter3D_hxx

#include <iostream>

#include "itkBinaryPruningImageFilter3D.h"
#include "itkImageRegionIterator.h"
#include "itkConstantBoundaryCondition.h"

namespace itk
{
	/**
	*    Constructor
	*/
	template< class TInputImage, class TOutputImage >
	BinaryPruningImageFilter3D< TInputImage, TOutputImage >
		::BinaryPruningImageFilter3D()
	{
		this->SetNumberOfRequiredOutputs(1);

		OutputImagePointer pruneImage = OutputImageType::New();
		this->SetNthOutput( 0, pruneImage.GetPointer() );

		m_Iteration =3;
	}

	/**
	*  Return the pruning Image pointer
	*/
	template< class TInputImage, class TOutputImage >
	typename BinaryPruningImageFilter3D<
		TInputImage, TOutputImage >::OutputImageType *
		BinaryPruningImageFilter3D< TInputImage, TOutputImage >
		::GetPruning(void)
	{
		return dynamic_cast< OutputImageType * >(
			this->ProcessObject::GetOutput(0) );
	}

	/**
	*  Prepare data for computation
	*/
	template< class TInputImage, class TOutputImage >
	void
		BinaryPruningImageFilter3D< TInputImage, TOutputImage >
		::ComputePruneImage(void)
	{
		OutputImagePointer pruneImage = GetPruning();

		InputImagePointer inputImage  =
			dynamic_cast< const TInputImage  * >( ProcessObject::GetInput(0) );


		pruneImage->SetBufferedRegion( pruneImage->GetRequestedRegion() );
		pruneImage->Allocate();

		typename OutputImageType::RegionType region  = pruneImage->GetRequestedRegion();

		ImageRegionConstIterator< TInputImage > it(inputImage,  region);
		ImageRegionIterator< TOutputImage >     ot(pruneImage,  region);

		typename NeighborhoodIteratorType::RadiusType radius;
		radius.Fill(1);
		ConstNeighborhoodIterator< InputImageType > itNeighbor(radius, inputImage, region);
		ConstantBoundaryCondition< InputImageType > nbc;
		itNeighbor.OverrideBoundaryCondition(&nbc);

		itNeighbor.GoToBegin();
		it.GoToBegin();
		ot.GoToBegin();

		const unsigned int neighborhoodSize = itNeighbor.Size();
		const unsigned int medianPosition = neighborhoodSize / 2;

		//First round: raw branch counting through neighbors
		while ( !it.IsAtEnd() )
		{
			//// collect all the pixels in the neighborhood, note that we use
			//// GetPixel on the NeighborhoodIterator to honor the boundary conditions
			int numBranch = 0;
			for ( unsigned int i = 0; i < neighborhoodSize; ++i )
			{
				if(itNeighbor.GetPixel(i) > 0.5)
					numBranch ++;
			}

			if(it.Get() < 0.5)
				ot.Set(0);
			else
			{
				ot.Set(numBranch-1);
			}
			++itNeighbor;
			++it;
			++ot;
		}

		//Second round: pruning
		NeighborhoodIterator< OutputImageType > itOutNeighbor(radius, pruneImage, region);
		ConstantBoundaryCondition< OutputImageType > nbcOut;
		itOutNeighbor.OverrideBoundaryCondition(&nbcOut);
		itOutNeighbor.GoToBegin();
		ot.GoToBegin();
		it.GoToBegin();
		std::vector<Index<3> > toPrune;
		while ( !it.IsAtEnd() )
		{
			// for evety end point
			if(abs(1-ot.Get())<1e-6)
			{
				toPrune.push_back(ot.GetIndex());
				itOutNeighbor.SetLocation(ot.GetIndex());

				unsigned int i = 0; 
				while ( i < neighborhoodSize && toPrune.size() <= m_Iteration + 1 )  
					// if the fiber segment is longer than m_iteration, stop processing this fiber
				{
					if( i!=medianPosition && (toPrune.size()<2 || toPrune[toPrune.size()-2]!= itOutNeighbor.GetIndex(i)) )
					{
						if(abs( 2-itOutNeighbor.GetPixel(i))<1e-6)
						{
							toPrune.push_back(itOutNeighbor.GetIndex(i));
							itOutNeighbor.SetLocation(itOutNeighbor.GetIndex(i));	
							i = 0;
						}
						// end of the fiber segment (encounting a branch point or the end point)
						else if(itOutNeighbor.GetPixel(i) >2.5 || abs(1-itOutNeighbor.GetPixel(i))<1e-6)
						{				
							// prune the fiber segment stored in toPrune vector 
							for (std::vector<Index<3> >::size_type j = 0;j != toPrune.size();j++)
							{
								ot.SetIndex(toPrune[j]);
								ot.Set(0);
							}
							// if it meet end point in the end, also prune this end point
							if( abs(1-itOutNeighbor.GetPixel(i))<1e-6 )
							{
								ot.SetIndex(itOutNeighbor.GetIndex(i));
								ot.Set(0);
							}

							break;
						}
						else 
							i++;
					}
					else 
						i++;
				}

				toPrune.clear();
			}

			ot.SetIndex(it.GetIndex());
			++it;
			++ot;

		}

		// Third round: set the non-zero point to 1
		ot.GoToBegin();
		while(!ot.IsAtEnd())
		{
			ot.Get()>0.5?ot.Set(1):ot.Set(0);
			++ot;
		}
	}

	/**
	*  Generate PruneImage
	*/
	template< class TInputImage, class TOutputImage >
	void
		BinaryPruningImageFilter3D< TInputImage, TOutputImage >
		::GenerateData()
	{
		this->ComputePruneImage();
	} // end GenerateData()

	/**
	*  Print Self
	*/
	template< class TInputImage, class TOutputImage >
	void
		BinaryPruningImageFilter3D< TInputImage, TOutputImage >
		::PrintSelf(std::ostream & os, Indent indent) const
	{
		Superclass::PrintSelf(os, indent);

		os << indent << "Pruning image: " << std::endl;
		os << indent << "Iteration: " << m_Iteration << std::endl;
	}
} // end namespace itk

#endif
