/*=========================================================================

Program:   Visualization Toolkit
Module:    vtkFiberFractionOverZ.cxx

Copyright (c) Xiaojie Zhao
All rights reserved.

This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notice for more information.

xiaojie:
This is a modification of vtkExtractHistogram.cxx.
So some names might seem not to be that appropriate here. 
e.g.we use *bin*, but we plot linechart finally. But it won't 
affect understanding.

=========================================================================*/
#include "vtkFiberFractionOverZ.h"

#include "vtkCellData.h"
#include "vtkCompositeDataIterator.h"
#include "vtkCompositeDataSet.h"
#include "vtkDataSet.h"
#include "vtkDoubleArray.h"
#include "vtkGraph.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkIntArray.h"
#include "vtkIOStream.h"
#include "vtkMath.h"
#include "vtkObjectFactory.h"
//#include "vtkOnePieceExtentTranslator.h"
#include "vtkPointData.h"
#include "vtkSmartPointer.h"
#include "vtkStreamingDemandDrivenPipeline.h"
#include "vtkTable.h"
//xiaojie
#include "vtkImageData.h"

#include <vector>
#include <map>
#include <string>

struct vtkEHInternals
{
	vtkEHInternals() : FieldAssociation(-1) {}
	struct ArrayValuesType
	{
		// The total of the values per bin - the second vector
		// is for arrays with multiple components
		std::vector<std::vector<double> > TotalValues;
	};
	typedef std::map<std::string, ArrayValuesType> ArrayMapType;
	ArrayMapType ArrayValues;
	int FieldAssociation;
};

vtkStandardNewMacro(vtkFiberFractionOverZ);
//-----------------------------------------------------------------------------
vtkFiberFractionOverZ::vtkFiberFractionOverZ() :
Component(0),
BinCount(10)
{
	this->SetInputArrayToProcess(
		0,
		0,
		0,
		vtkDataObject::FIELD_ASSOCIATION_POINTS_THEN_CELLS,//MODIFIED//TOCHECK
		vtkDataSetAttributes::SCALARS);
	this->Internal = new vtkEHInternals;
	this->UseCustomBinRanges = false;
	this->CustomBinRanges[0] = 0;
	this->CustomBinRanges[1] = 100;
}

//-----------------------------------------------------------------------------
vtkFiberFractionOverZ::~vtkFiberFractionOverZ()
{
	delete this->Internal;
}

//-----------------------------------------------------------------------------
void vtkFiberFractionOverZ::PrintSelf(ostream& os, vtkIndent indent)
{
	this->Superclass::PrintSelf(os,indent);

	os << indent << "Component: " << this->Component << "\n";
	os << indent << "BinCount: " << this->BinCount << "\n";
	os << indent << "UseCustomBinRanges: " << this->UseCustomBinRanges << endl;
	os << indent << "CustomBinRanges: " <<
		this->CustomBinRanges[0] << ", " << this->CustomBinRanges[1] << endl;
}

//-----------------------------------------------------------------------------
int vtkFiberFractionOverZ::FillInputPortInformation (int port,
													 vtkInformation *info)
{
	this->Superclass::FillInputPortInformation(port, info);

	info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkDataObject");
	return 1;
}

//-----------------------------------------------------------------------------
int vtkFiberFractionOverZ::GetInputFieldAssociation()
{
	vtkInformationVector *inArrayVec =
		this->Information->Get(INPUT_ARRAYS_TO_PROCESS());
	vtkInformation *inArrayInfo = inArrayVec->GetInformationObject(0);
	return inArrayInfo->Get(vtkDataObject::FIELD_ASSOCIATION());
}


bool vtkFiberFractionOverZ::InitializeBinExtents(
	vtkInformationVector** inputVector,
	vtkDoubleArray* bin_extents,
	double& min, double& max)
{
	double range[2];
	range[0] = VTK_DOUBLE_MAX;
	range[1] = -VTK_DOUBLE_MAX;

	// Keeping the column name constant causes less issues in the GUI.
	bin_extents->SetName("bin_extents");

	//obtain a pointer to the name of the vtkDataArray to bin up
	//and find the range of the data values within it

	vtkInformation* inInfo = inputVector[0]->GetInformationObject(0);
	vtkDataObject *input = inInfo->Get(vtkDataObject::DATA_OBJECT());
	vtkCompositeDataSet *cdin = vtkCompositeDataSet::SafeDownCast(input);
	if (cdin)
	{
		vtkErrorMacro("cannot process composite datasets");
	}
	else
	{
		vtkImageData *input_data = vtkImageData::SafeDownCast(input);
		range[0] = input_data->GetOrigin()[2] - input_data->GetSpacing()[2]/2;
		range[1] = input_data->GetOrigin()[2] - input_data->GetSpacing()[2]/2 
			+ input_data->GetSpacing()[2]*input_data->GetDimensions()[2];
	}

	if (this->UseCustomBinRanges)
	{
		range[0] = this->CustomBinRanges[0];
		range[1] = this->CustomBinRanges[1];
	}

	// Calculate the extents of each bin, based on the range of values in the
	// input ...
	if (range[0] == range[1])
	{
		// Give it some width.
		range[1] = range[0]+1;
		//vtkErrorMacro("min_z cannot be the same as max_z");
	}

	min = range[0];
	max = range[1];
	this->FillBinExtents(bin_extents, min, max);
	return true;
}

//-----------------------------------------------------------------------------
//NOMODIFY
void vtkFiberFractionOverZ::FillBinExtents(vtkDoubleArray* bin_extents,
										   double min, double max)
{
	// Calculate the extents of each bin, based on the range of values in the
	// input ...
	if (min == max)
	{
		// Give it some width.
		max = min + 1;
	}

	bin_extents->SetNumberOfComponents(1);
	bin_extents->SetNumberOfTuples(this->BinCount);
	double bin_delta = (max - min) / this->BinCount;
	double half_delta = bin_delta/2.0;
	for(int i = 0; i < this->BinCount; ++i)
	{
		bin_extents->SetValue(i, min + (i * bin_delta) + half_delta);
	}
}


inline int vtkFiberFractionOverZClamp(int value, int min, int max)
{
	value = value < min ? min : value;
	value = value > max ? max : value;
	return value;
}

//-----------------------------------------------------------------------------

void vtkFiberFractionOverZ::BinAnArray(vtkImageData* input_data,
									   vtkDoubleArray *bin_values,
									   double min, double max)
{
	// If the requested component is out-of-range for the input,
	// the bin_values will be 0, so no need to do any actual counting.
	if(input_data == NULL ||
		this->Component < 0 ||
		this->Component >= input_data->GetNumberOfScalarComponents())
	{
		return;
	}

	int dim[3];
	input_data->GetDimensions(dim);
	double* fiberFractionZ = new double[dim[2]];
	for(int z_idx = 0; z_idx < dim[2]; z_idx++)
	{
		//int allvoxel_num = 0;
		int fiber_num = 0;
		for (int x_idx = 0; x_idx < dim[0]; x_idx++)
			for(int y_idx = 0; y_idx < dim[1]; y_idx++)
				if(abs(input_data->GetScalarComponentAsDouble(x_idx, y_idx, z_idx, this->Component)) > 1e-6)//TOCHECK
					fiber_num++;
		fiberFractionZ[z_idx] = double(fiber_num) / (dim[0] * dim[1]);
	}



	//int num_of_tuples = data_array->GetNumberOfTuples();
	double bin_delta = (max-min)/this->BinCount;
	for(int bin_idx = 0; bin_idx < BinCount; bin_idx++)
	{
		double fiber_weight = 0;
		double bin_lower = bin_delta * bin_idx;
		double bin_upper = bin_delta * (bin_idx + 1);
		for(int z_idx = 0; z_idx < dim[2]; z_idx++)
		{
			double voxel_lower = input_data->GetSpacing()[2] * z_idx;
			double voxel_upper = input_data->GetSpacing()[2] * (z_idx + 1);
			fiber_weight += std::max(std::min(bin_upper, voxel_upper) - 
				std::max(bin_lower, voxel_lower),0.0) * fiberFractionZ[z_idx];
		}
		bin_values->SetValue(bin_idx, fiber_weight/bin_delta);
	}
	delete[] fiberFractionZ;
}

//-----------------------------------------------------------------------------
int vtkFiberFractionOverZ::RequestData(vtkInformation* /*request*/,
									   vtkInformationVector** inputVector,
									   vtkInformationVector* outputVector)
{

	//// Build an empty output grid in advance, so we can bail-out if we
	//// encounter any problems
	vtkTable* const output_data = vtkTable::GetData(outputVector, 0);
	output_data->Initialize();

	if (this->UseCustomBinRanges &&
		this->CustomBinRanges[1] < this->CustomBinRanges[0])
	{
		double min = this->CustomBinRanges[1];
		double max = this->CustomBinRanges[0];
		this->CustomBinRanges[0] = min;
		this->CustomBinRanges[1] = max;
		vtkWarningMacro("Custom bin range adjusted to keep min <= max value");
	}

	// These are the mid-points for each of the bins
	vtkSmartPointer<vtkDoubleArray> bin_extents =
		vtkSmartPointer<vtkDoubleArray>::New();
	bin_extents->SetNumberOfComponents(1);
	bin_extents->SetNumberOfTuples(this->BinCount);
	bin_extents->SetName("bin_extents");
	bin_extents->FillComponent(0, 0.0);

	// Insert values into bins ...
	vtkSmartPointer<vtkDoubleArray> bin_values =
		vtkSmartPointer<vtkDoubleArray>::New();
	bin_values->SetNumberOfComponents(1);
	bin_values->SetNumberOfTuples(this->BinCount);
	bin_values->SetName("bin_values");
	bin_values->FillComponent(0, 0.0);

	// Initializes the bin_extents array.
	double min, max;
	if (!this->InitializeBinExtents(inputVector, bin_extents, min, max))
	{
		this->Internal->ArrayValues.clear();
		return 1;
	}

	vtkInformation* inInfo = inputVector[0]->GetInformationObject(0);
	vtkDataObject *input = inInfo->Get(vtkDataObject::DATA_OBJECT());
	vtkCompositeDataSet *cdin = vtkCompositeDataSet::SafeDownCast(input);
	if (cdin)
	{
		vtkErrorMacro("cannot process composite datasets");
	}
	else
	{
		vtkImageData *input_data = vtkImageData::SafeDownCast(input);
		this->BinAnArray(input_data, bin_values, min, max);
	}

	output_data->GetRowData()->AddArray(bin_extents);
	output_data->GetRowData()->AddArray(bin_values);

	return 1;
}
