/*=========================================================================

Program:   Visualization Toolkit
Module:    vtkFiberFractionOverZ.h

Copyright (c) Xiaojie Zhao
All rights reserved.

This software is distributed WITHOUT ANY WARRANTY; without even
the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
// .NAME vtkFiberFractionOverZ - plot the fiber fraction over the distance from
// the cell surface. (Z) "bin number" is given as a parameter which means the 
// number of thin slices the whole Z is divided into.
// It is modified from the vtkExtractHistogram filter. The output vtkPolyData
// will have contain a vtkDoubleArray named "bin_extents" which contains
// the boundaries between each bin, and a vtkUnsignedLongArray named "bin_values" 
// which will contain the fiberfraction for each bin.

#ifndef __vtkFiberFractionOverZ_h
#define __vtkFiberFractionOverZ_h

#include "vtkTableAlgorithm.h"
//xiaojie
#include "vtkImageData.h"

//BTX
class vtkDoubleArray;
class vtkFieldData;
class vtkIntArray;
struct vtkEHInternals;
//ETX

class VTK_EXPORT vtkFiberFractionOverZ : public vtkTableAlgorithm
{
public:
	static vtkFiberFractionOverZ* New();
	vtkTypeMacro(vtkFiberFractionOverZ, vtkTableAlgorithm);
	void PrintSelf(ostream& os, vtkIndent indent);

	// Description:
	// Controls which input data component should be binned, for input arrays
	// with more-than-one component
	vtkSetClampMacro(Component, int, 0, VTK_INT_MAX);
	vtkGetMacro(Component, int);

	// Description:
	// Controls the number of bins N in the output
	vtkSetClampMacro(BinCount, int, 1, VTK_INT_MAX);
	vtkGetMacro(BinCount, int);

	// Description:
	// Get/Set custom bin ranges to use. These are used only when
	// UseCustomBinRanges is set to true.
	vtkSetVector2Macro(CustomBinRanges, double);
	vtkGetVector2Macro(CustomBinRanges, double);

	// Description:
	// When set to true, CustomBinRanges will  be used instead of using the full
	// range for the selected array. By default, set to false.
	vtkSetMacro(UseCustomBinRanges, bool);
	vtkGetMacro(UseCustomBinRanges, bool);
	vtkBooleanMacro(UseCustomBinRanges, bool);

protected: 
	vtkFiberFractionOverZ();
	~vtkFiberFractionOverZ();


	virtual int FillInputPortInformation (int port, vtkInformation *info);

	virtual int RequestData(vtkInformation *request, 
		vtkInformationVector **inputVector, 
		vtkInformationVector *outputVector);

	// Initialize the bin_extents using the data range for the selected 
	// array.
	virtual bool InitializeBinExtents(
		vtkInformationVector** inputVector, 
		vtkDoubleArray* bin_extents, 
		double& min, double& max);

	void BinAnArray(vtkImageData* input_data,
		vtkDoubleArray *bin_values,
		double min, double max);
	void FillBinExtents(vtkDoubleArray* bin_extents, double min, double max);

	double CustomBinRanges[2];
	bool UseCustomBinRanges;
	int Component;
	int BinCount;

	vtkEHInternals* Internal;

private:
	void operator=(const vtkFiberFractionOverZ&); // Not implemented
	vtkFiberFractionOverZ(const vtkFiberFractionOverZ&); // Not implemented

	int GetInputFieldAssociation();
	vtkFieldData* GetInputFieldData(vtkDataObject* input);
};

#endif
