#include <itkLandweberDeconvolutionImageFilter.h>
#include <vtkITKLandweberDeconvolutionImageFilter.h>

#include <functional>

#ifdef _WIN32
 #include <type_traits>
#else
 #ifdef _LIBCPP_VERSION
 // using libc++
 #include <type_traits>
 #else
 // using libstdc++
 #include <tr1/type_traits>
 #endif
#endif

#include <sitkBoundaryConditions.hxx>

#include <vtkDataArray.h>
#include <vtkObjectFactory.h>
#include <vtkImageData.h>
#include <vtkInformationVector.h>
#include <vtkInformation.h>
#include <vtkPointData.h>
#include <vtkStreamingDemandDrivenPipeline.h>

#include <itkImageToVTKImageFilter.h>
#include <itkVTKImageToImageFilter.h>

typedef unsigned char uint8_t;
typedef unsigned int  uint32_t;

#define ITK_UCHAR_TYPE
#define ITK_CHAR_TYPE
#define ITK_USHORT_TYPE
#define ITK_SHORT_TYPE
#define ITK_UINT_TYPE
#define ITK_INT_TYPE
#define ITK_ULONG_TYPE
#define ITK_LONG_TYPE
#define ITK_FLOAT_TYPE
#define ITK_DOUBLE_TYPE

vtkStandardNewMacro(vtkITKLandweberDeconvolutionImageFilter);

vtkITKLandweberDeconvolutionImageFilter::vtkITKLandweberDeconvolutionImageFilter()
{
  this->SetNumberOfInputPorts(2);
  this->SetNumberOfOutputPorts(1);

  this->Alpha = 0.1;
  this->NumberOfIterations = 1;
  this->Normalize = false ? 1 : 0;
  this->BoundaryCondition = ZERO_FLUX_NEUMANN_PAD;
  this->OutputRegionMode = SAME;
  this->SetInputArrayToProcess(0,0,0,vtkDataObject::FIELD_ASSOCIATION_POINTS,
    vtkDataSetAttributes::SCALARS);
  this->SetInputArrayToProcess(1,1,0,vtkDataObject::FIELD_ASSOCIATION_POINTS,
    vtkDataSetAttributes::SCALARS);
}

vtkITKLandweberDeconvolutionImageFilter::~vtkITKLandweberDeconvolutionImageFilter()
{
}

int vtkITKLandweberDeconvolutionImageFilter::RequestUpdateExtent(
  vtkInformation * vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector * vtkNotUsed(outputVector))
{
  int inWExt[6];
  vtkInformation *inInfo0 = inputVector[0]->GetInformationObject(0);
  inInfo0->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), inWExt);
  inInfo0->Set(vtkStreamingDemandDrivenPipeline::UPDATE_EXTENT(), inWExt, 6);

  vtkInformation *inInfo1 = inputVector[1]->GetInformationObject(0);
  inInfo1->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), inWExt);
  inInfo1->Set(vtkStreamingDemandDrivenPipeline::UPDATE_EXTENT(), inWExt, 6);

  return 1;
}

int vtkITKLandweberDeconvolutionImageFilter::FillInputPortInformation(int port, vtkInformation * info)
{
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkImageData");
  info->Set(vtkAlgorithm::INPUT_IS_REPEATABLE(), 0);
  info->Set(vtkAlgorithm::INPUT_IS_OPTIONAL(), 0);

  if (port == 0 || port == 1)
    {
    return 1;
    }

  return 0;
}

int vtkITKLandweberDeconvolutionImageFilter::RequestInformation(vtkInformation *vtkNotUsed(request), vtkInformationVector **inputVector, vtkInformationVector *outputVector)
{
  double spacing[3];
  double origin[3];
  int wholeExt[6];

  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  inInfo->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), wholeExt);
  inInfo->Get(vtkDataObject::SPACING(), spacing);
  inInfo->Get(vtkDataObject::ORIGIN(), origin);

  outInfo->Set(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(), wholeExt, 6);
  outInfo->Set(vtkDataObject::SPACING(), spacing, 3);
  outInfo->Set(vtkDataObject::ORIGIN(), origin, 3);

  return 1;
}

int vtkITKLandweberDeconvolutionImageFilter::RequestData(vtkInformation * vtkNotUsed(info), vtkInformationVector **inputVector, vtkInformationVector *outputVector)
{
  vtkInformation *inInfo0 = inputVector[0]->GetInformationObject(0);
  vtkImageData *input0 = vtkImageData::SafeDownCast
    (inInfo0->Get(vtkDataObject::DATA_OBJECT()));
  if (!input0)
    {
    vtkErrorMacro("Input 0 is not of type vtkImageData");
    return 0;
    }
  
  vtkInformation *inInfo1 = inputVector[1]->GetInformationObject(0);
  vtkImageData *input1 = vtkImageData::SafeDownCast
    (inInfo1->Get(vtkDataObject::DATA_OBJECT()));
  if (!input1)
    {
    vtkErrorMacro("Input 1 is not of type vtkImageData");
    return 0;
    }
  
  vtkInformation *outInfo = outputVector->GetInformationObject(0);
  vtkImageData *output = vtkImageData::SafeDownCast
    (outInfo->Get(vtkDataObject::DATA_OBJECT()));
  if (!output)
    {
    vtkErrorMacro("Output is not of type vtkImageData");
    return 0;
    }
  
  // Set the active scalar array in the inputs. Normally, this would be
  // done in the filter that imports the VTK image into ITK, but that
  // filter doesn't expose a way to set the active scalar in the interface.
  char * activeArrayName0 = this->GetInputArrayToProcess( 0, input0 )->GetName();
  input0->GetPointData()->SetActiveScalars( activeArrayName0 );
  char * activeArrayName1 = this->GetInputArrayToProcess( 1, input1 )->GetName();
  input1->GetPointData()->SetActiveScalars( activeArrayName1 );
  
  bool success = true;
  switch ( input0->GetScalarType() )
    {
  #ifdef ITK_UCHAR_TYPE
    case VTK_UNSIGNED_CHAR:
      success = this->Run( static_cast< unsigned char * >( 0 ), static_cast< unsigned char * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_CHAR_TYPE
    case VTK_CHAR:
    case VTK_SIGNED_CHAR:
      success = this->Run( static_cast< char * >( 0 ), static_cast< char * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_USHORT_TYPE
    case VTK_UNSIGNED_SHORT:
      success = this->Run( static_cast< unsigned short * >( 0 ), static_cast< unsigned short * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_SHORT_TYPE
    case VTK_SHORT:
      success = this->Run( static_cast< short * >( 0 ), static_cast< short * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_UINT_TYPE
    case VTK_UNSIGNED_INT:
      success = this->Run( static_cast< unsigned int * >( 0 ), static_cast< unsigned int * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_INT_TYPE
    case VTK_INT:
      success = this->Run( static_cast< int * >( 0 ), static_cast< int * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_ULONG_TYPE
    case VTK_UNSIGNED_LONG:
      success = this->Run( static_cast< unsigned long * >( 0 ), static_cast< unsigned long * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_LONG_TYPE
    case VTK_LONG:
      success = this->Run( static_cast< long * >( 0 ), static_cast< long * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_FLOAT_TYPE
    case VTK_FLOAT:
      success = this->Run( static_cast< float * >( 0 ), static_cast< float * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  #ifdef ITK_DOUBLE_TYPE
    case VTK_DOUBLE:
      success = this->Run( static_cast< double * >( 0 ), static_cast< double * >( 0 ),
                           input0, input1, output );
      break;
  #endif
  
    default:
      vtkErrorMacro( "Unknown pixel type encountered in input0" );
      success = false;
      break;
    }
  return (success ? 1 : 0);
}

template< class TInputPixel, class TInputPixel1 >
bool vtkITKLandweberDeconvolutionImageFilter::Run(TInputPixel *, TInputPixel1 *, vtkImageData * vtkInput0, vtkImageData * vtkInput1, vtkImageData * output)
{
  typedef itk::Image< TInputPixel, 3 > InputImageType;
  typedef InputImageType InputImageType2;
  typedef InputImageType OutputImageType;

  typedef itk::LandweberDeconvolutionImageFilter< InputImageType, InputImageType2,  OutputImageType > FilterType;
  typename FilterType::Pointer filter = FilterType::New();

  typedef itk::VTKImageToImageFilter< InputImageType > VTKToITKConnectorType0;
  typename VTKToITKConnectorType0::Pointer vtkToITKConnector0 = VTKToITKConnectorType0::New();
  vtkToITKConnector0->SetInput( vtkInput0 );
  try
    {
    vtkToITKConnector0->Update();
    }
  catch ( itk::ExceptionObject & except )
    {
    vtkErrorMacro( << except );
    return false;
    }

  typedef itk::VTKImageToImageFilter< InputImageType2 > VTKToITKConnectorType1;
  typename VTKToITKConnectorType1::Pointer vtkToITKConnector1 = VTKToITKConnectorType1::New();
  vtkToITKConnector1->SetInput( vtkInput1 );
  try
    {
    vtkToITKConnector1->Update();
    }
  catch ( itk::ExceptionObject & except )
    {
    vtkErrorMacro( << except );
    return false;
    }

  InputImageType * image1 = vtkToITKConnector0->GetOutput();
  InputImageType2 * image2 = vtkToITKConnector1->GetOutput();
  filter->SetInput( image1 ); filter->SetKernelImage( image2 );

  filter->SetAlpha(this->GetAlpha());
  filter->SetNumberOfIterations(this->GetNumberOfIterations());
  filter->SetNormalize(this->GetNormalize() != 0);

  std::auto_ptr< itk::ImageBoundaryCondition< InputImageType > > bc( itk::simple::CreateNewBoundaryConditionInstance< Self, FilterType >( BoundaryCondition ) ); filter->SetBoundaryCondition( bc.get() );

  filter->SetOutputRegionMode( typename FilterType::OutputRegionModeType( int( OutputRegionMode ) ) );
  try
    {
    filter->Update();
    }
  catch ( itk::ExceptionObject & except )
    {
    vtkErrorMacro( << except );
    return false;
    }

  typedef itk::ImageToVTKImageFilter< OutputImageType > ITKToVTKConnectorType;
  typename ITKToVTKConnectorType::Pointer itkToVTKConnector = ITKToVTKConnectorType::New();
  itkToVTKConnector->SetInput( filter->GetOutput() );
  itkToVTKConnector->Update();

  output->DeepCopy( itkToVTKConnector->GetOutput() );

  return true;
}

void vtkITKLandweberDeconvolutionImageFilter::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);

  os << indent << "Alpha:" << this->Alpha << "\n";
  os << indent << "NumberOfIterations:" << this->NumberOfIterations << "\n";
  os << indent << "Normalize:" << this->Normalize << "\n";
  os << indent << "BoundaryCondition:" << this->BoundaryCondition << "\n";
  os << indent << "OutputRegionMode:" << this->OutputRegionMode << "\n";
}
